"""
File: /__init__.py
Created Date: Tuesday July 22nd 2025
Author: Christian Nonis <alch.infoemail@gmail.com>
-----
Last Modified: Tuesday July 22nd 2025 11:40:55 am
Modified By: the developer formerly known as Christian Nonis at <alch.infoemail@gmail.com>
-----
"""

from .async_driver import AsyncLumenBrainDriver
from .sync_driver import LumenBrainDriver

__all__ = ["AsyncLumenBrainDriver", "LumenBrainDriver"]



"""
File: /__init__.py
Created Date: Saturday August 23rd 2025
Author: Christian Nonis <alch.infoemail@gmail.com>
-----
Last Modified: Saturday August 23rd 2025 2:45:17 pm
Modified By: the developer formerly known as Christian Nonis at <alch.infoemail@gmail.com>
-----
"""



"""
File: /async_driver.py
Created Date: Tuesday July 22nd 2025
Author: Christian Nonis <alch.infoemail@gmail.com>
-----
Last Modified: Tuesday July 22nd 2025 12:32:18 pm
Modified By: the developer formerly known as Christian Nonis at <alch.infoemail@gmail.com>
-----
"""

import asyncio
from typing import List, Literal, Optional

import aiohttp
from .constants.endpoints import (
    API_KEY_HEADER,
    MEMORY_CONTENT_TYPES,
    MemoryEndpoints,
    MemoryQueryResponse,
    MemoryUpdateResponse,
    InfoRetrievalResult,
)


class AsyncLumenBrainDriver:
    def __init__(self, api_key: str):
        if not api_key:
            raise ValueError("Lumen Brain API key is required")

        self.api_key = api_key

    async def save_message(
        self,
        memory_uuid: str,
        content: str,
        role: Optional[Literal["user", "assistant"]] = None,
        conversation_id: Optional[str] = None,
        metadata: Optional[dict] = None,
    ) -> MemoryUpdateResponse:
        """
        Request body for the memory update endpoint.

        Args:
            memory_uuid: The UUID of the memory to update, you can provide yours or let the API generate one.
            content: The text content of the message to save.
            role: Literal["user", "assistant"].
            conversation_id: The optional ID of the current conversation, if not provided a new conversation will be created.
            metadata: The optional metadata to add to the memory.
        """

        task_id = None
        conversation_id = None
        memory_id = None

        async with aiohttp.ClientSession() as session:
            try:
                async with session.post(
                    url=MemoryEndpoints.UPDATE.value,
                    headers={API_KEY_HEADER: self.api_key},
                    json={
                        "memory_uuid": memory_uuid,
                        "type": "message",
                        "content": content,
                        "role": role,
                        "conversation_id": conversation_id,
                        "metadata": metadata,
                    },
                ) as response:
                    result = await response.json()
                    task_id = result.get("task_id")
                    conversation_id = result.get("conversation_id")
                    memory_id = result.get("memory_id")
            except Exception as e:
                print("[LUMEN BRAIN] Error saving message", e)
                raise e

            if not conversation_id or memory_id:
                return {
                    "error": "Failed to save message",
                    "task_id": task_id,
                }

            if result.get("error"):
                return {
                    "status": "error",
                    "error": result.get("error"),
                }
            return MemoryUpdateResponse(**result)

    async def inject_knowledge(
        self,
        memory_uuid: str,
        content: str,
        resource_type: Optional[MEMORY_CONTENT_TYPES] = None,
        metadata: Optional[dict] = None,
    ) -> MemoryUpdateResponse:
        """
        Request body for the memory update endpoint.

        Args:
            memory_uuid: The UUID of the memory to update, you can provide yours or let the API generate one.
            content: The text content of the message to save.
            resource_type: Literal["file", "event", "webpage", "email"].
            metadata: The optional metadata to add to the memory.
        """
        task_id = None
        conversation_id = None
        memory_id = None

        max_retries = 3
        base_delay = 1

        for attempt in range(max_retries):
            try:
                async with aiohttp.ClientSession() as session:
                    async with session.post(
                        url=MemoryEndpoints.UPDATE.value,
                        headers={API_KEY_HEADER: self.api_key},
                        json={
                            "memory_uuid": memory_uuid,
                            "type": resource_type,
                            "content": content,
                            "resource_type": resource_type,
                            "metadata": metadata,
                        },
                    ) as response:
                        if response.status in [502, 503, 504]:
                            error_text = await response.text()
                            if attempt < max_retries - 1:
                                delay = base_delay * (2**attempt)
                                print(
                                    f"[LUMEN BRAIN] Retry {attempt + 1}/{max_retries} after {delay}s due to status {response.status}"
                                )
                                await asyncio.sleep(delay)
                                continue
                            raise Exception(
                                f"API returned status {response.status} after {max_retries} attempts: {error_text[:200]}"
                            )

                        if response.status != 200:
                            error_text = await response.text()
                            raise Exception(
                                f"API returned status {response.status}: {error_text[:200]}"
                            )

                        content_type = response.headers.get("Content-Type", "")
                        if "application/json" not in content_type:
                            error_text = await response.text()
                            raise Exception(
                                f"Expected JSON response but got {content_type}: {error_text[:200]}"
                            )

                        result = await response.json()
                        task_id = result.get("task_id")
                        memory_id = result.get("memory_id")
                        conversation_id = result.get("conversation_id")

                        if not conversation_id or not memory_id:
                            return {
                                "error": "Failed to inject knowledge",
                                "task_id": task_id,
                            }

                        if result.get("error"):
                            return {
                                "status": "error",
                                "error": result.get("error"),
                            }
                        return MemoryUpdateResponse(**result)

            except (aiohttp.ClientError, asyncio.TimeoutError) as e:
                if attempt < max_retries - 1:
                    delay = base_delay * (2**attempt)
                    print(
                        f"[LUMEN BRAIN] Retry {attempt + 1}/{max_retries} after {delay}s due to network error: {e}"
                    )
                    await asyncio.sleep(delay)
                    continue
                print(
                    f"[LUMEN BRAIN] Error injecting knowledge after {max_retries} attempts",
                    e,
                )
                raise e
            except Exception as e:
                print("[LUMEN BRAIN] Error injecting knowledge", e)
                raise e

    async def query_memory(
        self, text: str, memory_uuid: str, conversation_id: str
    ) -> MemoryQueryResponse:
        """
        Request body for the memory query endpoint.

        Args:
            text: The text to query the memory with (usually the same message as the one that was sent to the agent).
            memory_uuid: The UUID of the memory to query (usually it's related to a user).
            conversation_id: The optional ID of the current conversation, if not provided a new conversation will be created.
        """
        async with aiohttp.ClientSession() as session:
            async with session.post(
                url=MemoryEndpoints.QUERY.value,
                headers={API_KEY_HEADER: self.api_key},
                json={
                    "text": text,
                    "memory_uuid": memory_uuid,
                    "conversation_id": conversation_id,
                },
            ) as response:
                result = await response.json()
                try:
                    if result.get("error"):
                        return {
                            "status": "error",
                            "error": result.get("error"),
                        }
                    return MemoryQueryResponse(**result)
                except Exception as e:
                    print("[LUMEN BRAIN] Error querying memory", e)
                    raise e

    async def fetch_info(
        self, memory_uuid: str, entities: List[str], info: str, depth: int = 2
    ) -> InfoRetrievalResult:
        """
        Fetch information and relationships about entities in the memory

        Args:
            memory_uuid: The UUID of the memory to query (usually it's related to a user).
            entities: The entities that are related to the information to be retrieved.
            info: The information to be retrieved.
            depth: The higher relation depth that will be looked for.
        """
        if not memory_uuid:
            raise ValueError("Memory UUID is required")
        if not entities:
            raise ValueError("Entities are required and must be a list of strings")
        if not info:
            raise ValueError("Info is required and must be a string")

        async with aiohttp.ClientSession() as session:
            async with session.get(
                url=MemoryEndpoints.QUERY_ENTITIES.value,
                headers={API_KEY_HEADER: self.api_key},
                params={
                    "memory_uuid": memory_uuid,
                    "entities": entities,
                    "info": info,
                    "depth": depth,
                },
            ) as response:
                result = await response.json()
                try:
                    if result.get("error"):
                        return {
                            "status": "error",
                            "error": result.get("error"),
                        }
                    return InfoRetrievalResult(**result)
                except Exception as e:
                    print("[LUMEN BRAIN] Error fetching info", e)
                    raise e



"""
File: endpoints.py
Created Date: Tuesday July 22nd 2025
Author: Christian Nonis <alch.infoemail@gmail.com>
"""

from enum import Enum
from typing import List, Literal, Optional, Union
from pydantic import BaseModel, Field, field_validator

API_VERSION = "v1"
BASE_URL = f"https://api.brainapi.lumen-labs.ai/api/{API_VERSION}"

MEMORY_CONTENT_TYPES = Literal[
    "file",
    "file_chunk",
    "event",
    "webpage",
    "webpage_chunk",
    "email",
    "email_chunk",
]
ALL_MEMORY_CONTENT_TYPES = Union[Literal["message"], MEMORY_CONTENT_TYPES]

API_KEY_HEADER = "X-LumenBrain-ApiKey"


class MemoryEndpoints(str, Enum):
    UPDATE = f"{BASE_URL}/memory/update"
    QUERY = f"{BASE_URL}/memory/query"
    QUERY_ENTITIES = f"{BASE_URL}/memory/query/entities"
    TASKS = f"{BASE_URL}/tasks"


class MemoryQueryResponse(BaseModel):
    """
    Response for the memory query endpoint.

    Args:
        context: The textual context relevant for the query.
    """

    context: str


class MemoryUpdateResponse(BaseModel):
    task_id: str
    memory_id: Optional[str] = None
    conversation_id: Optional[str] = None
    error: Optional[str] = None


class ApiHeaders(BaseModel):
    api_key: str = Field(alias="X-LumenBrain-ApiKey")


class Relationship(BaseModel):
    name: str
    phrase: str
    direction: str
    natural_lang: str
    related_node: "Node"
    similarity: float
    source_id: str


class Node(BaseModel):
    natural_lang: str
    id: str
    parent_id: Optional[str] = None
    name: str
    relationships: Optional[List[Relationship]] = []
    source_id: str
    phrase: str

    @field_validator("relationships", mode="before")
    @classmethod
    def parse_relationships(cls, v):
        """
        Parse the relationships field from the API response.
        """
        if isinstance(v, str):
            if v == "[]":
                return []
            import json

            try:
                parsed = json.loads(v)
                return parsed if isinstance(parsed, list) else []
            except (json.JSONDecodeError, TypeError):
                return []
        return v


class InfoRetrievalResult(BaseModel):
    nodes: List[Node]
    most_relevant_relation: Optional[Relationship] = None
    most_relevant_confidence: Optional[float] = None



"""
File: /sync_driver.py
Created Date: Tuesday July 22nd 2025
Author: Christian Nonis <alch.infoemail@gmail.com>
-----
Last Modified: Tuesday July 22nd 2025 12:32:11 pm
Modified By: the developer formerly known as Christian Nonis at <alch.infoemail@gmail.com>
-----
"""

import time
from typing import List, Literal, Optional

import requests
from .constants.endpoints import (
    ALL_MEMORY_CONTENT_TYPES,
    API_KEY_HEADER,
    MemoryEndpoints,
    MemoryQueryResponse,
    MemoryUpdateResponse,
    InfoRetrievalResult,
)


class LumenBrainDriver:
    def __init__(self, api_key: str):
        if not api_key:
            raise ValueError("Lumen Brain API key is required")

        self.api_key = api_key

    def save_message(
        self,
        memory_uuid: str,
        type: ALL_MEMORY_CONTENT_TYPES,
        content: str,
        role: Optional[Literal["user", "assistant"]] = None,
        conversation_id: Optional[str] = None,
        metadata: Optional[dict] = None,
    ) -> MemoryUpdateResponse:
        """
        Request body for the memory update endpoint.

        Args:
            memory_uuid: The UUID of the memory to update, you can provide yours or let the API generate one.
            content: The text content of the message to save.
            role: Literal["user", "assistant"].
            conversation_id: The optional ID of the current conversation, if not provided a new conversation will be created.
            metadata: The optional metadata to add to the memory.
        """
        task_id = None

        try:
            response = requests.post(
                url=MemoryEndpoints.UPDATE.value,
                headers={API_KEY_HEADER: self.api_key},
                json={
                    "memory_id": memory_uuid,
                    "type": type,
                    "content": content,
                    "role": role,
                    "conversation_id": conversation_id,
                    "metadata": metadata,
                },
            )
            result = response.json()
            task_id = result.get("task_id")
        except Exception as e:
            print("[LUMEN BRAIN] Error saving message", e)
            raise e

        result = None

        while not result:
            try:
                result_res = requests.get(
                    url=f"{MemoryEndpoints.TASKS.value}/{task_id}",
                    headers={API_KEY_HEADER: self.api_key},
                )
                if result_res.status_code == 200:
                    result = result_res.json()
                    break
            except Exception as e:
                print("[LUMEN BRAIN] Error polling task", e)
                raise e
            time.sleep(1)

        if result.get("error"):
            return {
                "status": "error",
                "error": result.get("error"),
            }
        return MemoryUpdateResponse(**result)

    def inject_knowledge(
        self,
        memory_uuid: str,
        type: ALL_MEMORY_CONTENT_TYPES,
        content: str,
        conversation_id: Optional[str] = None,
        metadata: Optional[dict] = None,
    ) -> MemoryUpdateResponse:
        """
        Request body for the memory update endpoint.

        Args:
            memory_uuid: The UUID of the memory to update, you can provide yours or let the API generate one.
            content: The text content of the message to save.
            resource_type: Literal["file", "event", "webpage", "email"].
            conversation_id: The optional ID of the current conversation, if not provided a new conversation will be created.
            metadata: The optional metadata to add to the memory.
        """
        task_id = None

        try:
            response = requests.post(
                url=MemoryEndpoints.UPDATE.value,
                headers={API_KEY_HEADER: self.api_key},
                json={
                    "memory_id": memory_uuid,
                    "type": type,
                    "content": content,
                    "resource_type": type,
                    "conversation_id": conversation_id,
                    "metadata": metadata,
                },
            )
            result = response.json()
            task_id = result.get("task_id")
        except Exception as e:
            print("[LUMEN BRAIN] Error injecting knowledge", e)
            raise e

        result = None

        while not result:
            try:
                result_res = requests.get(
                    url=f"{MemoryEndpoints.TASKS.value}/{task_id}",
                    headers={API_KEY_HEADER: self.api_key},
                )
                if result_res.status_code == 200:
                    result = result_res.json()
                    break
            except Exception as e:
                print("[LUMEN BRAIN] Error polling knowledge injection task", e)
                raise e
            time.sleep(1)

        if result.get("error"):
            return {
                "status": "error",
                "error": result.get("error"),
            }
        return MemoryUpdateResponse(**result)

    def query_memory(
        self, text: str, memory_uuid: str, conversation_id: str
    ) -> MemoryQueryResponse:
        """
        Request body for the memory query endpoint.

        Args:
            text: The text to query the memory with (usually the same message as the one that was sent to the agent).
            memory_uuid: The UUID of the memory to query (usually it's related to a user).
            conversation_id: The optional ID of the current conversation, if not provided a new conversation will be created.
        """
        response = requests.post(
            url=MemoryEndpoints.QUERY.value,
            headers={API_KEY_HEADER: self.api_key},
            json={
                "text": text,
                "memory_id": memory_uuid,
                "conversation_id": conversation_id,
            },
        )
        result = response.json()
        if result.get("error"):
            return {
                "status": "error",
                "error": result.get("error"),
            }
        return MemoryQueryResponse(**result)

    def fetch_info(
        self, memory_uuid: str, entities: List[str], info: str, depth: int = 2
    ) -> InfoRetrievalResult:
        """
        Fetch information and relationships about entities in the memory

        Args:
            memory_uuid: The UUID of the memory to query (usually it's related to a user).
            entities: The entities that are related to the information to be retrieved.
            info: The information to be retrieved.
            depth: The higher relation depth that will be looked for.
        """

        if not memory_uuid:
            raise ValueError("Memory UUID is required")
        if not entities:
            raise ValueError("Entities are required and must be a list of strings")
        if not info:
            raise ValueError("Info is required and must be a string")

        try:
            response = requests.get(
                url=MemoryEndpoints.QUERY_ENTITIES.value,
                headers={API_KEY_HEADER: self.api_key},
                params={
                    "memory_uuid": memory_uuid,
                    "entities": entities,
                    "info": info,
                    "depth": depth,
                },
            )
        except Exception as e:
            print("[LUMEN BRAIN] Error fetching info", e)
            raise e

        result = response.json()
        if result.get("error"):
            return {
                "status": "error",
                "error": result.get("error"),
            }
        return InfoRetrievalResult(**result)
